local ADDON, Package = ...
local Color = Package.Color
local ColorPalette = Package.ColorPalette

local BACKDROP_DIALOG_32_16_4444 = {
  bgFile = "Interface\\DialogFrame\\UI-DialogBox-Background",
  edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Border",
  tile = true,
  tileEdge = true,
  tileSize = 32,
  edgeSize = 16,
  insets = {left = 4, right = 4, top = 4, bottom = 4}
}

local function escapable(frame)
  table.insert(UISpecialFrames, frame:GetName())
end

local function movable(source, target)
  local function Source_OnDragStart(self)
    target:StartMoving()
  end

  local function Source_OnDragStop(self)
    target:StopMovingOrSizing()
  end

  source:RegisterForDrag("LeftButton")
  source:EnableMouse(true)

  source:SetScript("OnDragStart", Source_OnDragStart)
  source:SetScript("OnDragStop", Source_OnDragStop)

  target:SetMovable(true)
end

local function create_window(name, parent)
  local frame = CreateFrame("Frame", name, parent, "BackdropTemplate")
  frame:Hide()

  frame:SetSize(544, 360)
  frame:SetPoint("CENTER")
  frame:SetBackdrop(BACKDROP_DIALOG_32_16_4444)
  frame:SetToplevel(true)
  frame:SetFrameStrata("DIALOG")

  local close_button = CreateFrame("Button", "$parentCloseButton", frame, "UIPanelCloseButton")
  close_button:SetSize(24, 24)
  close_button:SetPoint("TOPRIGHT", -2, -2)

  return frame
end

local function create_tab_button(name, parent)
  local frame = CreateFrame("Button", name, parent)
  frame:SetSize(64, 28)
  frame:SetFrameStrata("LOW")

  local left = frame:CreateTexture(nil, "BACKGROUND")
  left:SetSize(16, 28)
  left:SetPoint("TOPLEFT")
  left:SetTexture([[Interface\ChatFrame\ChatFrameTab]])
  left:SetTexCoord(0.0, 0.25, 0.0, 1.0)

  local middle = frame:CreateTexture(nil, "BACKGROUND")
  middle:SetSize(44, 28)
  middle:SetPoint("LEFT", left, "RIGHT")
  middle:SetTexture([[Interface\ChatFrame\ChatFrameTab]])
  middle:SetTexCoord(0.25, 0.75, 0.0, 1.0)

  local right = frame:CreateTexture(nil, "BACKGROUND")
  right:SetSize(16, 28)
  right:SetPoint("LEFT", middle, "RIGHT")
  right:SetTexture([[Interface\ChatFrame\ChatFrameTab]])
  right:SetTexCoord(0.75, 1.0, 0.0, 1.0)

  local highlight = frame:CreateTexture(nil, "BACKGROUND")
  highlight:SetPoint("TOPLEFT", left, 2, -4)
  highlight:SetPoint("BOTTOMRIGHT", right, -2, -8)
  highlight:SetTexture([[Interface\PaperDollInfoFrame\UI-Character-Tab-Highlight]])
  highlight:SetBlendMode("ADD")

  local pushed = frame:CreateTexture(nil, "BACKGROUND")
  pushed:SetPoint("TOPLEFT", left, 0, -2)
  pushed:SetPoint("BOTTOMRIGHT", right, -2, -8)
  pushed:SetTexture([[Interface\PaperDollInfoFrame\UI-Character-Tab-Highlight-yellow]])
  pushed:SetBlendMode("ADD")

  local font_string = frame:CreateFontString(nil, "ARTWORK", "GameFontNormalSmall")
  font_string:SetPoint("LEFT", left, 8, -6)
  font_string:SetPoint("RIGHT", right, -8, -6)
  font_string:SetJustifyH("CENTER")

  frame:SetHighlightTexture(highlight)
  frame:SetPushedTextOffset(0, 0)
  frame:SetPushedTexture(pushed)
  frame:SetFontString(font_string)

  return frame
end

local function create_tab_pane(name, parent)
  local frame = CreateFrame("Frame", name, parent)

  return frame
end

local function create_tab(name, parent, items)
  local frame = CreateFrame("Frame", "$parent" .. name, parent)
  frame:SetPoint("BOTTOMLEFT", parent, "TOPLEFT", 0, -2)
  frame:SetPoint("BOTTOMRIGHT", parent, "TOPRIGHT", 0, 0)
  frame:SetHeight(28)

  local function Button_OnClick(self)
    local parent = self:GetParent()
    local id = self:GetID()

    for at = 1, #items do
      local button = _G[parent:GetName() .. "Button" .. at]
      local pane = _G[parent:GetName() .. "Pane" .. at]

      button:SetButtonState("NORMAL", false)
      pane:Hide()

      if at == id then
        button:SetButtonState("PUSHED", true)
        pane:Show()
      end
    end
  end

  local relative_to = nil

  for at = 1, #items do
    local button = create_tab_button("$parentButton" .. at, frame)
    button:SetID(at)
    button:SetText(items[at])
    button:SetScript("OnClick", Button_OnClick)

    if relative_to then
      button:SetPoint("TOPLEFT", relative_to, "TOPRIGHT", 6, 0)
    else
      button:SetPoint("TOPLEFT", 2, 0)
    end

    local pane = create_tab_pane("$parentPane" .. at, frame)
    pane:SetID(at)
    pane:SetAllPoints(parent)

    relative_to = button
  end

  local function Select(self, index)
    local button = _G[self:GetName() .. "Button" .. index]
    local script = button:GetScript("OnClick")

    script(button)
  end

  frame.Select = Select

  return frame
end

local function create_listbox(name, parent)
  local frame = CreateFrame("ScrollFrame", name, parent, "UIPanelScrollFrameTemplate")
  frame.ScrollBar:ClearAllPoints()
  frame.ScrollBar:SetPoint("TOPLEFT", frame, "TOPRIGHT", 2, -12)
  frame.ScrollBar:SetPoint("BOTTOMLEFT", frame, "BOTTOMRIGHT", 0, 12)

  frame.ScrollBar.ThumbTexture:SetSize(14, 18)
  frame.ScrollBar.ScrollUpButton:SetSize(14, 12)
  frame.ScrollBar.ScrollDownButton:SetSize(14, 12)

  local function OnVerticalScroll(self, offset)
    ScrollFrame_OnVerticalScroll(self, offset)

    self.item_height = #self.items > 0 and math.floor(self.items[1]:GetHeight()) or 0
    self.items_offset = math.floor((offset / math.max(1, self.item_height)) + 0.5)
  end

  local function OnScrollRangeChanged(self, xrange, yrange)
    ScrollFrame_OnScrollRangeChanged(self, xrange, math.max(0, (#self.data - #self.items) * self.item_height))

    self.ScrollBar:SetObeyStepOnDrag(true)
    self.ScrollBar:SetValueStep(self.item_height)
    self.ScrollBar:SetStepsPerPage(#self.items - 1)
  end

  local function ForceUpdate(self)
    local OnVerticalScroll = self:GetScript("OnVerticalScroll")
    OnVerticalScroll(self, self.ScrollBar:GetValue())

    local OnScrollRangeChanged = self:GetScript("OnScrollRangeChanged")
    OnScrollRangeChanged(self, 0, 0)
  end

  frame:SetScript("OnVerticalScroll", OnVerticalScroll)
  frame:SetScript("OnScrollRangeChanged", OnScrollRangeChanged)

  frame.ForceUpdate = ForceUpdate

  return frame
end

local function create_searchbox(name, parent)
  local frame = CreateFrame("EditBox", name, parent, "SearchBoxTemplate")
  frame:SetWidth(120)
  frame:SetHeight(24)

  return frame
end

local function create_checkbox(name, parent)
  local frame = CreateFrame("CheckButton", name, parent, "UICheckButtonTemplate")

  return frame
end

local function create_trade_character_card(name, parent)
  local frame = CreateFrame("Button", nil, parent)
  local texture = frame:CreateTexture(nil, "ARTWORK")
  local font_string = frame:CreateFontString(nil, "ARTWORK", "GameFontNormalSmall")
  frame:SetSize(16 + 4 + 104, 16)
  frame:SetFontString(font_string)
  frame:SetPushedTextOffset(0, 0)
  frame:SetNormalTexture(texture)
  texture:ClearAllPoints()
  texture:SetSize(16, 16)
  texture:SetPoint("LEFT")
  texture:SetTexCoord(0.08, 0.92, 0.08, 0.92)
  font_string:SetJustifyH("LEFT")
  font_string:SetPoint("TOPLEFT", texture, "TOPRIGHT", 4, 0)
  font_string:SetPoint("BOTTOMRIGHT", frame)

  return frame
end

local function create_trade_character_money(name, parent)
  local font_string = parent:CreateFontString(nil, "ARTWORK", "GameFontNormalSmall")
  font_string:SetSize(120, 16)
  font_string:SetJustifyH("RIGHT")
  font_string:SetVertexColor(1, 1, 1)

  return font_string
end

local function create_trade_character_items(name, parent)
  local function OnEnter(self)
    if self.item_id then
      GameTooltip:SetOwner(self)
      GameTooltip:SetItemByID(self.item_id)
      GameTooltip:Show()
    end
  end

  local function OnLeave(self)
    GameTooltip:Hide()
  end

  local stash = {}

  for id = 1, MAX_TRADE_ITEMS do
    local frame = CreateFrame("Button", name, parent)
    frame:SetID(id)
    frame:SetSize(30, 30)
    frame:SetNormalTexture([[Interface\Buttons\UI-Quickslot2]])
    frame:GetNormalTexture():SetTexCoord(0.2, 0.8, 0.2, 0.8)
    frame:SetHighlightTexture([[Interface\Buttons\ButtonHilight-Square]], "ADD")
    frame:SetScript("OnEnter", OnEnter)
    frame:SetScript("OnLeave", OnLeave)

    if TRADE_ENCHANT_SLOT == id then
      frame:GetNormalTexture():SetVertexColor(0.1, 1, 0.1)
    end

    local icon = frame:CreateTexture(nil, "ARTWORK", nil, -8)
    icon:SetAllPoints(frame)

    local quantity = frame:CreateFontString(nil, "ARTWORK", "NumberFont_Shadow_Tiny")
    quantity:SetPoint("TOPLEFT", frame, 0, 0)
    quantity:SetPoint("BOTTOMRIGHT", frame, -2, 1)
    quantity:SetJustifyH("RIGHT")
    quantity:SetJustifyV("BOTTOM")
    quantity:SetTextColor(1, 1, 1)

    local quality_border = frame:CreateTexture(nil, "ARTWORK", nil, -7)
    quality_border:SetTexture([[Interface\Buttons\UI-ActionButton-Border]])
    quality_border:SetTexCoord(0.2, 0.8, 0.2, 0.8)
    quality_border:SetAllPoints(frame)
    quality_border:SetBlendMode("ADD")

    frame.icon = icon
    frame.quantity = quantity
    frame.quality_border = quality_border

    table.insert(stash, frame)
  end

  return stash
end

local function create_trade_listbox_item(name, parent)
  local frame = CreateFrame("Button", name, parent)
  frame:Hide()
  frame:SetPoint("LEFT")
  frame:SetPoint("RIGHT")
  frame:SetHeight(64)

  local id = frame:CreateFontString(nil, "ARTWORK", "GameFontNormalTiny2")
  id:SetSize(102, 14)
  id:SetPoint("TOPLEFT")
  id:SetJustifyH("LEFT")
  id:SetVertexColor(0.75, 0.61, 0)

  local timestamp = frame:CreateFontString(nil, "ARTWORK", "GameFontNormalTiny2")
  timestamp:SetSize(102, 14)
  timestamp:SetPoint("TOPRIGHT")
  timestamp:SetJustifyH("RIGHT")
  timestamp:SetVertexColor(0.75, 0.61, 0)

  local map = CreateFrame("Button", nil, frame)
  local map_font_string = map:CreateFontString(nil, "ARTWORK", "GameFontNormalTiny2")
  map:SetPoint("TOPLEFT", id, "TOPRIGHT")
  map:SetPoint("BOTTOMRIGHT", timestamp, "BOTTOMLEFT")
  map:SetFontString(map_font_string)
  map:SetPushedTextOffset(0, 0)
  map_font_string:SetJustifyH("CENTER")
  map_font_string:SetAllPoints(map)
  map_font_string:SetVertexColor(1, 1, 1)

  local player_card = create_trade_character_card(nil, frame)
  player_card:SetPoint("TOPLEFT", id, "BOTTOMLEFT")

  local player_items = create_trade_character_items(nil, frame)

  local player_money = create_trade_character_money(nil, frame)
  player_money:SetPoint("LEFT", player_card, "RIGHT", 0, 0)

  local player_background = frame:CreateTexture(nil, "BACKGROUND")
  player_background:SetTexture([[Interface\Buttons\WHITE8X8]])
  player_background:SetPoint("TOPLEFT", player_card, "BOTTOMLEFT")
  player_background:SetPoint("BOTTOMRIGHT", player_money, 2, -1)
  player_background:SetVertexColor(0.75, 0.61, 0)

  local target_card = create_trade_character_card(nil, frame)
  target_card:SetPoint("TOPLEFT", frame, 266, -14) -- TODO

  local target_items = create_trade_character_items(nil, frame)

  local target_money = create_trade_character_money(nil, frame)
  target_money:SetPoint("LEFT", target_card, "RIGHT", 0, 0)

  local target_background = frame:CreateTexture(nil, "BACKGROUND")
  target_background:SetTexture([[Interface\Buttons\WHITE8X8]])
  target_background:SetPoint("TOPLEFT", target_card, "BOTTOMLEFT")
  target_background:SetPoint("BOTTOMRIGHT", target_money, 2, -1)
  target_background:SetVertexColor(0.75, 0.61, 0)

  local player_relative_to = nil
  local target_relative_to = nil

  for id = 1, MAX_TRADE_ITEMS do
    local player_item = player_items[id]
    local target_item = target_items[id]

    if player_relative_to and target_relative_to then
      player_item:SetPoint("LEFT", player_relative_to, "RIGHT", 6, 0)
      target_item:SetPoint("LEFT", target_relative_to, "RIGHT", 6, 0)
    else
      player_item:SetPoint("TOPLEFT", player_background, "BOTTOMLEFT", 0, -2)
      target_item:SetPoint("TOPLEFT", target_background, "BOTTOMLEFT", 0, -2)
    end

    player_relative_to = player_item
    target_relative_to = target_item
  end

  frame.id = id
  frame.map = map
  frame.timestamp = timestamp
  frame.player_card = player_card
  frame.player_money = player_money
  frame.player_items = player_items
  frame.target_card = target_card
  frame.target_money = target_money
  frame.target_items = target_items

  return frame
end

local function update_trade_listbox()
  local state = MedusaTradeStorage
  local search_term = MedusaTradeHistoryWindowTabPane1SearchBox:GetText():trim()
  local search_results = {}

  if search_term:len() > 0 then
    for _, entry in pairs(state.history) do
      local map_zone = entry.map_zone or ""
      local map_subzone = entry.map_subzone or ""
      local player_name = entry.player_name or ""
      local player_guild = entry.player_guild or ""
      local target_name = entry.target_name or ""
      local target_guild = entry.target_guild or ""

      if map_zone:find(search_term)
      or map_subzone:find(search_term)
      or player_name:find(search_term)
      or player_guild:find(search_term)
      or target_name:find(search_term)
      or target_guild:find(search_term) then
        table.insert(search_results, entry)
      end

      for at = 1, MAX_TRADE_ITEMS do
        local player_item = entry.player_items[at]
        local target_item = entry.target_items[at]

        if (player_item and player_item.name:find(search_term))
        or (target_item and target_item.name:find(search_term))
        or (player_item and player_item.enchantment and player_item.enchantment:find(search_term))
        or (target_item and target_item.enchantment and target_item.enchantment:find(search_term)) then
          table.insert(search_results, entry)
          break
        end
      end
    end
  end

  if #search_results > 0 or search_term:len() > 0 then
    MedusaTradeHistoryWindowTabPane1ListBox.data = search_results
  else
    MedusaTradeHistoryWindowTabPane1ListBox.data = state.history
  end

  MedusaTradeHistoryWindowTabPane1ListBox:ForceUpdate()
end

local function history()
  local tab1_name, option1_name = "", ""

  if GetLocale() == "koKR" then
    tab1_name = "거래 기록"
    option1_name = "거래시 내역을 귓속말"
  else
    tab1_name = "History"
    option1_name = "Send transaction details to target"
  end

  local window = create_window(ADDON .. "HistoryWindow", UIParent)
  window:SetScript("OnShow", function (self)
    local listbox = _G[ADDON .. "HistoryWindowTabPane1ListBox"]
    if listbox then
      listbox:ForceUpdate()
    end
  end)

  local tab = create_tab("Tab", window, {tab1_name, OPTIONS})
  tab:Select(1)

  local checkbox = create_checkbox("$parentWhisperCheckbox", _G[tab:GetName() .. "Pane2"])
  checkbox.text:SetText(option1_name)
  checkbox:SetSize(24, 24)
  checkbox:SetPoint("TOPLEFT", 8, -8)
  checkbox:SetScript("OnShow", function (self)
    if MedusaTradeStorage then
      self:SetChecked(MedusaTradeStorage.whisper)
    end
  end)
  checkbox:SetScript("OnClick", function (self)
    if MedusaTradeStorage then
      MedusaTradeStorage.whisper = self:GetChecked()
    end
  end)

  local searchbox = create_searchbox("$parentSearchBox", _G[tab:GetName() .. "Pane1"])
  searchbox:HookScript("OnTextChanged", function (self) update_trade_listbox() end)
  searchbox:SetPoint("TOP", 0, -6)

  local listbox = create_listbox("$parentListBox", _G[tab:GetName() .. "Pane1"])
  listbox:SetPoint("TOPLEFT", 8, -34)
  listbox:SetPoint("BOTTOMRIGHT", -24, 6)

  listbox.data = MedusaTradeStorage.history
  listbox.items = {}

  local relative_to = nil

  for at = 0, 4 do
    local item = create_trade_listbox_item(nil, listbox)

    if relative_to then
      item:SetPoint("TOP", relative_to, "BOTTOM")
    else
      item:SetPoint("TOP", listbox)
    end

    relative_to = item

    table.insert(listbox.items, item)
  end

  local empty_font_string = listbox:CreateFontString(nil, "ARTWORK", "GameFontNormal")
  empty_font_string:SetPoint("CENTER")
  empty_font_string:SetSize(200, 16)
  empty_font_string:SetJustifyH("CENTER")
  empty_font_string:SetJustifyV("MIDDLE")
  empty_font_string:SetText("No Results")
  empty_font_string:Hide()

  local function format_id(id)
    return string.format("%03d", id)
  end

  local function format_map(zone, subzone, pvp_type)
    return Color.format(ColorPalette[pvp_type and pvp_type:upper() or "NORMAL"], string.join(" - ", zone, subzone))
  end

  local function format_timestamp(timestamp)
    return date("%m-%d %H:%M:%S", timestamp)
  end

  local function format_character_card(name, class, level)
    return Color.format(GetQuestDifficultyColor(level), level) .. " " .. Color.format(RAID_CLASS_COLORS[class:upper()], name)
  end

  local function format_character_card_icon(sex, race)
    return GetRaceAtlas(race, sex)
  end

  local function format_character_money(money)
    return GetMoneyString(money, true)
  end

  local function ListBox_OnVerticalScroll(self, offset)
    for at, line in pairs(self.items) do
      local cursor = at + self.items_offset
      local entry = self.data[cursor]

      empty_font_string:SetShown(#self.data == 0)

      if cursor <= #self.data then
        line.id:SetText(format_id(cursor))
        line.map:SetText(format_map(entry.map_zone, entry.map_subzone, entry.map_pvp))
        line.timestamp:SetText(format_timestamp(entry.timestamp))

        line.player_card:SetText(format_character_card(entry.player_name, entry.player_class, entry.player_level))
        line.player_card:GetNormalTexture():SetAtlas(format_character_card_icon(entry.player_sex, entry.player_race))
        line.player_money:SetText(format_character_money(entry.player_money))

        line.target_card:SetText(format_character_card(entry.target_name, entry.target_class, entry.target_level))
        line.target_card:GetNormalTexture():SetAtlas(format_character_card_icon(entry.target_sex, entry.target_race))
        line.target_money:SetText(format_character_money(entry.target_money))

        for slot_id = 1, MAX_TRADE_ITEMS do
          line.player_items[slot_id].item_id = nil
          line.player_items[slot_id].item_enchantment = nil
          line.player_items[slot_id].icon:SetTexture([[Interface\PaperDoll\UI-PaperDoll-Slot-Bag]])
          line.player_items[slot_id].quantity:Hide()
          line.player_items[slot_id].quantity:SetText(1)
          line.player_items[slot_id].quality_border:Hide()
          line.player_items[slot_id].quality_border:SetVertexColor(1, 1, 1)

          local item = entry.player_items[slot_id]

          if item and item.icon and item.quality and item.quantity then
            line.player_items[slot_id].item_id = item.id
            line.player_items[slot_id].item_enchantment = item.enchantment
            line.player_items[slot_id].icon:SetTexture(item.icon)
            line.player_items[slot_id].quantity:SetText(item.quantity)
            line.player_items[slot_id].quantity:SetShown(item.quantity > 1)
            line.player_items[slot_id].quality_border:SetVertexColor(ITEM_QUALITY_COLORS[item.quality].color:GetRGBA())
            line.player_items[slot_id].quality_border:Show()
          end

          line.target_items[slot_id].item_id = nil
          line.target_items[slot_id].item_enchantment = nil
          line.target_items[slot_id].icon:SetTexture([[Interface\PaperDoll\UI-PaperDoll-Slot-Bag]])
          line.target_items[slot_id].quantity:Hide()
          line.target_items[slot_id].quantity:SetText(1)
          line.target_items[slot_id].quality_border:Hide()
          line.target_items[slot_id].quality_border:SetVertexColor(1, 1, 1)

          local item = entry.target_items[slot_id]

          if item and item.icon and item.quality and item.quantity then
            line.target_items[slot_id].item_id = item.id
            line.target_items[slot_id].item_enchantment = item.enchantment
            line.target_items[slot_id].icon:SetTexture(item.icon)
            line.target_items[slot_id].quantity:SetText(item.quantity)
            line.target_items[slot_id].quantity:SetShown(item.quantity > 1)
            line.target_items[slot_id].quality_border:SetVertexColor(ITEM_QUALITY_COLORS[item.quality].color:GetRGBA())
            line.target_items[slot_id].quality_border:Show()
          end
        end

        line:Show()
      else
        line:Hide()
      end
    end
  end

  listbox:HookScript("OnVerticalScroll", ListBox_OnVerticalScroll)
  listbox:ForceUpdate()

  local gametooltip_tooltip_set_item = function (self)
    local owner = self:GetOwner()

    if owner and owner.item_id and owner.item_enchantment then
      GameTooltip:AddLine(ITEM_PROPOSED_ENCHANT:format(owner.item_enchantment), 0.1, 1, 0.1)
    end
  end

  GameTooltip:HookScript("OnTooltipSetItem", gametooltip_tooltip_set_item)

  escapable(_G[ADDON .. "HistoryWindow"])
  movable(_G[ADDON .. "HistoryWindowTabButton1"], _G[ADDON .. "HistoryWindow"])
end

Package.history = history
