local ADDON, Package = ...
local Enum = Package.Enum

local Message = {}

function Message:format_name(name)
  return "<" .. name .. ">"
end

function Message:format_money(amount)
  local stash = ""

  local g = math.floor(amount / 10000)
  local s = math.floor((amount - (g * 10000)) / 100)
  local c = math.fmod(amount, 100)

  if g > 0 then
    stash = stash .. g .. GOLD_AMOUNT_SYMBOL .. " "
  end

  if s > 0 then
    stash = stash .. s .. SILVER_AMOUNT_SYMBOL .. " "
  end

  if c > 0 or (s == 0 and g == 0) then
    stash = stash .. c .. COPPER_AMOUNT_SYMBOL
  end

  return string.trim(stash)
end

function Message:format_item(name, quantity, enchantment)
  local stash = name

  if enchantment then
    stash = "(" .. stash .. "@" .. enchantment .. ")"
  else
    stash = "[" .. stash .. "]"
  end

  if quantity > 1 then
    stash = stash .. "x" .. tostring(quantity)
  end

  return stash
end

function Message:format(name, money, items)
  local stash = {}

  for _, item in pairs(items) do
    if item and item.name and item.quantity then
      table.insert(stash, self:format_item(item.name, item.quantity, item.enchantment))
    end
  end

  if money > 0 then
    table.insert(stash, 1, self:format_money(money))
  end

  if #stash > 0 then
    table.insert(stash, 1, self:format_name(name))
  end

  return stash
end

function Message:merge(items)
  local stash = {}
  local map = {}

  for at, item in ipairs(items) do
    if item and item.name and item.quantity then
      if map[item.name] then
        map[item.name].quantity = map[item.name].quantity + item.quantity
      else
        map[item.name] = item

        stash[at] = item
      end
    else
      stash[at] = item
    end
  end

  return stash
end

function Message:accept(transaction)
  local player = self:format(transaction.player_name, transaction.player_money, self:merge(transaction.player_items))
  local target = self:format(transaction.target_name, transaction.target_money, self:merge(transaction.target_items))

  local entries = Enum.concat(player, target)
  local payload = {}
  local content = ""

  for _, entry in ipairs(entries) do
    local ahead = content .. " " .. entry

    if string.len(ahead) > 251 then -- 4 bytes reserved
      table.insert(payload, string.trim(content))

      content = entry
    else
      content = ahead
    end
  end

  if string.len(content) > 0 then
    table.insert(payload, string.trim(content))
  end

  local length = #payload

  for at = 1, length do
    local content = nil

    if length > 1 then
      content = string.format("%d/%d %s", at, length, payload[at])
    else
      content = string.format("%s", payload[at])
    end

    SendChatMessage(content, "WHISPER", nil, transaction.target_name)
  end
end

function Message:cancel(transaction)
  SendChatMessage(ERR_TRADE_CANCELLED, "WHISPER", nil, transaction.target_name)
end

Package.Message = Message
